# Entry point for the EM Interpolation GUI
import sys
from PyQt5.QtWidgets import QApplication
from em_interp.gui.main_window import MainWindow


def main():
    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()
